! fe_model.f90 -- fixed effects binary choice model
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

! This module implements a basic fixed effects binary choice model.
! The following assumptions are made:
!
! * There are T = 2 time periods.
! * There are two covariates with the coefficient on the first normalized to 1.
! * The distribution of X is discrete and uniform.
! * The unobservables u_t are iid according to the standard Normal distribution.
! * The fixed effect is the average of the components of x's across time.
module fe_model
  use osl_core, wp => osl_wp
  use osl_dist
  use osl_random
  use dataset
  implicit none

  private

  public :: fe_model_t
  public :: fe_model_init
  public :: fe_model_free
  public :: fe_model_generate_c
  public :: fe_model_generate_data
  public :: fe_model_population_objective
  public :: fe_model_sample_objective

  integer, parameter :: dim_x = 2
  integer, parameter :: n_t = 2

  type :: fe_model_t
     real(wp) :: beta0                            ! True parameter
     integer :: n_x                               ! Cardinality of support of X
     real(wp), dimension(:,:), pointer :: x       ! Joint regressor support
     type(osl_discrete_rnd_t) :: xdist            ! Discrete X distribution
  end type fe_model_t

contains

  ! Initialize a new model and allocate memory.
  subroutine fe_model_init(self, x1, x2, beta0)
    type(fe_model_t), intent(inout) :: self
    real(wp), dimension(:), intent(in) :: x1
    real(wp), dimension(:), intent(in) :: x2
    real(wp), intent(in) :: beta0

    real(wp), dimension(:), allocatable :: px
    integer :: i, j, k, n_x1, n_x2

    ! Store parameters
    self%beta0 = beta0

    ! Store basic attributes
    n_x1 = size(x1)
    n_x2 = size(x2)
    self%n_x = n_x1 * n_x2
    allocate(self%x(2,self%n_x))
    allocate(px(self%n_x))

    ! Construct support of x
    i = 0
    do j = 1, n_x1
       do k = 1, n_x2
          i = i + 1
          self%x(1,i) = x1(j)
          self%x(2,i) = x2(k)
       end do
    end do

    ! Construct point masses for uniform X
    px = 1.0_wp / real(self%n_x, wp)

    ! Prepare to the discrete distribution of regressors
    call osl_discrete_rnd_init(self%xdist, px)

    ! Clean up
    deallocate(px)
  end subroutine fe_model_init

  ! Destroy an existing model, freeing any allocated memory.
  subroutine fe_model_free(self)
    type(fe_model_t), intent(inout) :: self
    
    ! Free the discrete distribution
    call osl_discrete_rnd_free(self%xdist)

    ! Free dynamically allocated memory
    deallocate(self%x)
  end subroutine fe_model_free

  ! Return the value of the individual effect given x_1 and x_2.
  function fe_model_generate_c(self, x1, x2) result(c)
    type(fe_model_t), intent(in) :: self
    real(wp), dimension(dim_x), intent(in) :: x1
    real(wp), dimension(dim_x), intent(in) :: x2
    real(wp) :: c

    c = sum(x1 + x2) / real(n_t * dim_x, wp)
  end function fe_model_generate_c

  ! Generate a sample from the model.
  subroutine fe_model_generate_data(self, rng, dat)
    type(fe_model_t), intent(inout) :: self
    type(osl_rng_t), intent(inout) :: rng
    type(dataset_t), intent(inout) :: dat
    integer :: ii, it, ix
    real(wp) :: xbeta, c
    real(wp), dimension(dat%t) :: u

    ! Sanity checks
    if (dat%t /= n_t) then
       stop 'The fixed effects model is only implemented for T = 2.'
    end if

    ! Generate data for each individual ii
    do ii = 1, dat%n

       ! Generate regressors and disturbances for each period it
       do it = 1, dat%t
          ! Regressors
          ix = osl_discrete_rnd(self%xdist, rng)
          dat%x(:,it,ii) = self%x(:,ix)

          ! Disturbances
          u(it) = osl_std_normal_rnd(rng)
       end do

       ! Draw the value of the fixed effect for individual ii
       c = fe_model_generate_c(self, dat%x(:,1,ii), dat%x(:,2,ii))

       ! Determine the choice in each period it
       do it = 1, dat%t
          xbeta = dat%x(1,it,ii) + self%beta0 * dat%x(2,it,ii)
          if (xbeta + c + u(it) > 0.0_wp) then
             dat%y(it,ii) = 1.0_wp
          else
             dat%y(it,ii) = 0.0_wp
          end if
       end do
    end do
  end subroutine fe_model_generate_data

  ! Evaluate the population objective function at theta
  subroutine fe_model_population_objective(self, beta, y)
    type(fe_model_t), intent(in) :: self
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: y

    integer :: i, j
    real(wp) :: idx1, idx2, pr_y1, pr_y2, c

    ! Calculate the expectation of (y2 - y1) * sgn((x2-x1)'beta),
    ! using the law of iterated expectations, conditioning on (x1, x2).
    y = 0.0_wp
    do i = 1, self%n_x
       do j = 1, self%n_x
          ! Determine the value of the fixed effect
          c = fe_model_generate_c(self, self%x(:,i), self%x(:,j))

          ! Evaluate the conditional choice probabilities using theta0
          idx1 = -self%x(1,i) - self%beta0 * self%x(2,i) - c
          idx2 = -self%x(1,j) - self%beta0 * self%x(2,j) - c

          pr_y1 = 1.0_wp - osl_std_normal_cdf(idx1)
          pr_y2 = 1.0_wp - osl_std_normal_cdf(idx2)

          ! Evaluate the sgn portion using theta
          idx1 = self%x(1,i) + beta * self%x(2,i)
          idx2 = self%x(1,j) + beta * self%x(2,j)

          ! Accumulate the value for this (x1, x2) pair
          y = y + (pr_y2 - pr_y1) * osl_sgn(idx2 - idx1)
       end do
    end do
    y = y / real(self%n_x * self%n_x, wp)
  end subroutine fe_model_population_objective

  ! Evaluate the sample objective function at theta
  subroutine fe_model_sample_objective(self, dat, beta, Q)
    type(fe_model_t), intent(in) :: self
    type(dataset_t), intent(in) :: dat
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q
    real(wp), dimension(dim_x) :: w
    integer :: i

    ! Calculate the sample average of (y2 - y1) * sgn((x2-x1)'beta).
    Q = 0.0_wp
    do i = 1, dat%n
       w = dat%x(:,2,i) - dat%x(:,1,i)
       Q = Q + (dat%y(2,i) - dat%y(1,i)) * osl_sgn(w(1) + beta * w(2))
    end do
    Q = Q / real(dat%n, wp)
  end subroutine fe_model_sample_objective

end module fe_model
