! feb_model.f90 -- fixed effects binary choice model with bounded regressors
!
! Copyright (C) 2009-2011 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

! This module implements a basic fixed effects binary choice model.
! The following assumptions are made:
!
! * There are T = 2 time periods.
! * There are two covariates with the coefficient on the first normalized to 1.
! * The distribution of the first component of X is uniform.
! * The second component of X is binary.
! * The unobservables u_t are iid according to the standard Normal distribution.
! * The fixed effect is the average of the components of x's across time.
module feb_model
  use osl_core, wp => osl_wp
  use osl_dist
  use osl_random
  use dataset
  implicit none

  private

  public :: n_t, dim_x
  public :: feb_model_t
  public :: feb_model_init
  public :: feb_model_free
  public :: feb_model_generate_c
  public :: feb_model_generate_data
  public :: feb_model_population_objective
  public :: feb_model_sample_objective

  integer, parameter :: dim_x = 2
  integer, parameter :: n_t = 2

  type :: feb_model_t
     ! True parameters
     real(wp) :: beta0                            ! True parameter

     ! X_{1t} is Uniform on X1_RANGE
     real(wp), dimension(2) :: x1_range           ! Upper and lower bounds for X_{1t}

     ! Simulation draws for X_{1t}
     integer :: n_x2                              ! Number of points in X_{2t} support
     real(wp), dimension(:), pointer :: x2_sup    ! Support of X_{2t}
     type(osl_discrete_rnd_t) :: x2_dist          ! Discrete X_{2t} distribution

     ! Simulation draws for X_{1t}
     integer :: nsim                              ! Number of Monte Carlo draws
     type(osl_rng_t) :: rng                       ! Random Number Generator
     real(wp), dimension(:,:), pointer :: x1_sim  ! Monte Carlo draws
  end type feb_model_t

contains

  ! Initialize a new model and allocate memory.
  subroutine feb_model_init(self, x1_range, x2_sup, beta0, nsim)
    type(feb_model_t), intent(inout) :: self
    real(wp), dimension(2), intent(in) :: x1_range
    real(wp), dimension(:), intent(in) :: x2_sup
    real(wp), intent(in) :: beta0
    integer, intent(in) :: nsim
    integer :: r, t
    real(wp), dimension(:), allocatable :: px2

    ! Store parameters
    self%beta0 = beta0

    ! Store bounds on support of X_{1t}
    self%x1_range = x1_range

    ! Store the support of X_{2t}
    self%n_x2 = size(x2_sup, 1)
    allocate(self%x2_sup(self%n_x2))
    self%x2_sup = x2_sup

    ! Set up discrete uniform distribution for X_2
    allocate(px2(self%n_x2))
    px2 = 1.0_wp / real(self%n_x2, wp)
    call osl_discrete_rnd_init(self%x2_dist, px2)

    ! Seed the PRNG used for simulation draws
    call osl_rng_seed(self%rng, 9842098)

    ! Simulation draws for X_{1t}
    self%nsim = nsim
    allocate(self%x1_sim(n_t, nsim))
    do r = 1, nsim
       do t = 1, n_t
          self%x1_sim(t, r) = osl_uniform_rnd(self%rng, &
               self%x1_range(1), self%x1_range(2))
       end do
    end do
  end subroutine feb_model_init

  ! Destroy an existing model, freeing any allocated memory.
  subroutine feb_model_free(self)
    type(feb_model_t), intent(inout) :: self

    deallocate(self%x1_sim)
    deallocate(self%x2_sup)
    call osl_discrete_rnd_free(self%x2_dist)
  end subroutine feb_model_free

  ! Return the value of the individual effect given x_1 and x_2.
  function feb_model_generate_c(self, x1, x2) result(c)
    type(feb_model_t), intent(in) :: self
    real(wp), dimension(dim_x), intent(in) :: x1
    real(wp), dimension(dim_x), intent(in) :: x2
    real(wp) :: c

    c = sum(x1 + x2) / real(n_t * dim_x, wp)
  end function feb_model_generate_c

  ! Generate a sample from the model.
  subroutine feb_model_generate_data(self, rng, dat)
    type(feb_model_t), intent(inout) :: self
    type(osl_rng_t), intent(inout) :: rng
    type(dataset_t), intent(inout) :: dat
    integer :: ii, it, ix2
    real(wp) :: xbeta, c
    real(wp), dimension(dat%t) :: u

    ! Sanity checks
    if (dat%t /= n_t) then
       stop 'The fixed effects model is only implemented for T = 2.'
    end if

    ! Generate data for each individual ii
    do ii = 1, dat%n

       ! Generate regressors and disturbances for each period it
       do it = 1, dat%t
          ! Regressors
          dat%x(1,it,ii) = osl_uniform_rnd(rng, self%x1_range(1), self%x1_range(2))
          ix2 = osl_discrete_rnd(self%x2_dist, rng)
          dat%x(2,it,ii) = self%x2_sup(ix2)

          ! Disturbances
          u(it) = osl_std_normal_rnd(rng)
       end do

       ! Draw the value of the fixed effect for individual ii
       c = feb_model_generate_c(self, dat%x(:,1,ii), dat%x(:,2,ii))

       ! Determine the choice in each period it
       do it = 1, dat%t
          xbeta = dat%x(1,it,ii) + self%beta0 * dat%x(2,it,ii)
          if (xbeta + c + u(it) > 0.0_wp) then
             dat%y(it,ii) = 1.0_wp
          else
             dat%y(it,ii) = 0.0_wp
          end if
       end do
    end do
  end subroutine feb_model_generate_data

  ! Evaluate the population objective function at theta
  subroutine feb_model_population_objective(self, beta, y)
    type(feb_model_t), intent(in) :: self
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: y

    integer :: i, j, r
    real(wp) :: idx1, idx2, pr_y1, pr_y2, c
    real(wp), dimension(2) :: x1, x2
    real(wp) :: frac

    ! Calculate the expectation of (y2 - y1) * sgn((x2-x1)'beta) via
    ! Monte Carlo integration.
    y = 0.0_wp
    frac = 1.0_wp / real(self%n_x2 * self%n_x2 * self%nsim, wp)
    do i = 1, self%n_x2
       do j = 1, self%n_x2
          do r = 1, self%nsim
             ! Covariates
             x1 = (/ self%x1_sim(1,r), self%x2_sup(i) /)
             x2 = (/ self%x1_sim(2,r), self%x2_sup(j) /)

             ! Determine the value of the fixed effect
             c = feb_model_generate_c(self, x1, x2)

             ! Evaluate the conditional choice probabilities using theta0
             idx1 = -x1(1) - self%beta0 * x1(2) - c
             idx2 = -x2(1) - self%beta0 * x2(2) - c

             pr_y1 = 1.0_wp - osl_std_normal_cdf(idx1)
             pr_y2 = 1.0_wp - osl_std_normal_cdf(idx2)

             ! Evaluate the sgn portion using theta
             idx1 = x1(1) + beta * x1(2)
             idx2 = x2(1) + beta * x2(2)

             ! Accumulate the value for this (x1, x2) pair
             y = y + (pr_y2 - pr_y1) * osl_sgn(idx2 - idx1)
          end do
       end do
    end do

    y = y * frac
  end subroutine feb_model_population_objective

  ! Evaluate the sample objective function at theta
  subroutine feb_model_sample_objective(self, dat, beta, Q)
    type(feb_model_t), intent(in) :: self
    type(dataset_t), intent(in) :: dat
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q
    real(wp), dimension(dim_x) :: w
    integer :: i

    ! Calculate the sample average of (y2 - y1) * sgn((x2-x1)'beta).
    Q = 0.0_wp
    do i = 1, dat%n
       w = dat%x(:,2,i) - dat%x(:,1,i)
       Q = Q + (dat%y(2,i) - dat%y(1,i)) * osl_sgn(w(1) + beta * w(2))
    end do
    Q = Q / real(dat%n, wp)
  end subroutine feb_model_sample_objective

end module feb_model
