! mcfeb.f90 -- fixed effects binary choice model Monte Carlo experiments
!
! Copyright (C) 2009-2011 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program mcfeb
  use osl_core, wp => osl_wp
  use osl_random
  use osl_stat, only: osl_sample_stats
  use osl_sort, only: osl_quicksort
  use dataset
  use feb_model
  use utility
  implicit none

  include 'revision.inc'

  !===== User Settings ========================================================!

  !----- Monte Carlo Parameters -----------------------------------------------!

  ! Number of Monte Carlo trials for each sample size, slackness, and level
  integer, parameter :: n_mc = 1000

  ! Sample sizes
  integer, parameter :: n_samp_sz = 8
  integer, parameter, dimension(n_samp_sz) :: samp_sz &
       = (/ 125, 250, 500, 1000, 2000, 4000, 8000, 16000 /)

  ! Slackness sequences (proportional to sqrt(log(n)))
  integer, parameter :: n_const = 4
  real(wp), dimension(n_const), parameter :: const &
       = (/ 0.20_wp, 0.10_wp, 0.05_wp, 0.00_wp /)

  ! Set DO_CR to .FALSE. to bypass calculation of confidence regions
  logical, parameter :: do_cr = .true.

  ! Confidence region coverage levels
  integer, parameter :: n_alpha = 3
  real(wp), parameter, dimension(n_alpha) :: alpha &
       = (/ 0.25_wp, 0.10_wp, 0.05_wp /)

  ! Subsampling (subsample size is n^(ss_frac_num / ss_frac_denom))
  integer, parameter :: n_ss = 100
  integer, parameter :: ss_frac_num = 3
  integer, parameter :: ss_frac_denom = 5

  !----- Model Specification --------------------------------------------------!

  ! Bounds on support of X_{1t}
  real(wp), parameter, dimension(2) :: x1_range = (/ 0.0_wp, 1.0_wp /)

  ! Finite support of X_{2t}
  real(wp), parameter, dimension(3) :: x2_sup = (/ 1.0_wp, 2.0_wp, 3.0_wp /)

  ! Population parameters
  real(wp), parameter :: beta0 = -0.6_wp

  ! Simulation
  integer, parameter :: nsim = 10000

  !----- Grid Search ----------------------------------------------------------!

  integer, parameter :: n_grid = 6001                      ! Grid size
  real(wp) :: grid_lb = -3.0_wp                            ! Lower grid bounds
  real(wp) :: grid_ub = 3.0_wp                             ! Upper grid bounds

  !----- Internal Variables ---------------------------------------------------!

  ! Evaluate Q and Q_N on a grid of size N_GRID.
  real(wp), dimension(n_grid) :: beta_grid                 ! Support points
  real(wp), dimension(n_grid) :: Q_pop                     ! Population obj. fn.

  ! Data
  integer :: n                                             ! Sample size
  integer :: m                                             ! Subsample size
  real(wp) :: ss_frac                                      ! Subsample fraction
  type(dataset_t), dimension(:), allocatable :: dat        ! Datasets
  type(dataset_t), dimension(:,:), allocatable :: dat_ss   ! Subsamples

  ! Population characteristics
  type(feb_model_t) :: mod                                  ! Model
  real(wp), dimension(2) :: id_set                         ! Identified set
  character(len=100) :: population                         ! Reporting

  ! Results
  real(wp), dimension(:,:,:), allocatable :: est
  real(wp), dimension(:,:,:,:), allocatable :: cr

  ! Statistics
  real(wp), dimension(:,:,:), allocatable :: est_mean
  real(wp), dimension(:,:), allocatable :: est_cover
  real(wp), dimension(:,:), allocatable :: est_dist
  real(wp), dimension(:,:,:), allocatable :: est_sd
  real(wp), dimension(:,:,:,:), allocatable :: cr_mean
  real(wp), dimension(:,:,:), allocatable :: cr_cover
  real(wp), dimension(:,:,:), allocatable :: cr_dist

  ! Random number generator
  type(osl_rng_t) :: rng

  ! Generic indices
  integer :: i, j, k, l, r

  !===== Main Program =========================================================!

  !----- Preparations ---------------------------------------------------------!

  ! Allocate memory
  allocate(dat(n_mc))
  allocate(est(2, n_const, n_mc))
  allocate(est_mean(2, n_const, n_samp_sz))
  allocate(est_cover(n_const, n_samp_sz))
  allocate(est_dist(n_const, n_samp_sz))
  allocate(est_sd(2, n_const, n_samp_sz))

  if (do_cr) then
     allocate(dat_ss(n_ss, n_mc))
     allocate(cr(2, n_alpha, n_const, n_mc))
     allocate(cr_mean(2, n_alpha, n_const, n_samp_sz))
     allocate(cr_cover(n_alpha, n_const, n_samp_sz))
     allocate(cr_dist(n_alpha, n_const, n_samp_sz))
  end if

  ! Seed the PRNG
  call osl_rng_seed(rng, 2480129)

  ! Initialize the fixed-effects model.
  call feb_model_init(mod, x1_range, x2_sup, beta0, nsim)

  ! Construct a grid for theta.
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)

  ! Subsample size
  ss_frac = real(ss_frac_num, wp) / real(ss_frac_denom, wp)

  !----- Determine \Theta_I ---------------------------------------------------!

  ! Obtain identified set by evaluating population objective function on grid.
  do k = 1, n_grid
     call feb_model_population_objective(mod, beta_grid(k), Q_pop(k))
  end do
  call get_max_set(beta_grid, Q_pop, 1, epsilon(1.0_wp), id_set)

  !----- Print Summary --------------------------------------------------------!

  call print_summary()

  !----- Monte Carlo Experiments ----------------------------------------------!

  ! For each sample size
  do i = 1, n_samp_sz
     call tic()
     n = samp_sz(i)                               ! Sample size
     m = int(real(n, wp)**ss_frac)                ! Subsample size
     call print_experiment()

     !$OMP PARALLEL DO
     do r = 1, n_mc
        ! Generate all data and subsamples in a critical region, using
        ! a common PRNG, so that the results will be reproducible when
        ! running in parallel.  Although the order in which datasets
        ! are generated and processed may differ, summary statistics
        ! will be unchanged.

        !$OMP CRITICAL
        call dataset_init(dat(r), n, n_t, dim_x)
        call feb_model_generate_data(mod, rng, dat(r))
        if (do_cr) then
           do j = 1, n_ss
              call dataset_init(dat_ss(j,r), m, n_t, dim_x)
              call subsample_dat(dat(r), m, dat_ss(j,r))
           end do
        end if
        !$OMP END CRITICAL

        ! Perform a single Monte Carlo trial and store the results.
        call run_experiment(dat(r), dat_ss(:,r), est(:,:,r), cr(:,:,:,r))

        ! Free memory
        call dataset_free(dat(r))
        if (do_cr) then
           do j = 1, n_ss
              call dataset_free(dat_ss(j,r))
           end do
        end if
     end do
     !$OMP END PARALLEL DO

     ! Average estimate, coverage, and mean Hausdorff distance.
     do l = 1, n_const
        call osl_sample_stats(est(:,l,:), MEAN=est_mean(:,l,i), SD=est_sd(:,l,i))
        est_cover(l,i) = 0.0_wp
        est_dist(l,i) = 0.0_wp
        do r = 1, n_mc
           if (cover_p(est(:,l,r), id_set)) then
              est_cover(l,i) = est_cover(l,i) + 1.0_wp
           end if
           est_dist(l,i) = est_dist(l,i) + hausdorff(est(:,l,r), id_set)
        end do
        est_cover(l,i) = est_cover(l,i) / real(n_mc, wp)
        est_dist(l,i) = est_dist(l,i) / real(n_mc, wp)
     end do

     ! Average confidence region, coverage, and Hausdorff distance.
     if (do_cr) then
        do l = 1, n_const
           do k = 1, n_alpha
              call osl_sample_stats(cr(:,k,l,:), MEAN=cr_mean(:,k,l,i))
              cr_cover(k,l,i) = 0.0_wp
              cr_dist(k,l,i) = 0.0_wp
              do r = 1, n_mc
                 if (cover_p(cr(:,k,l,r), id_set)) then
                    cr_cover(k,l,i) = cr_cover(k,l,i) + 1.0_wp
                 end if
                 cr_dist(k,l,i) = cr_dist(k,l,i) + hausdorff(cr(:,k,l,r), id_set)
              end do
              cr_cover(k,l,i) = cr_cover(k,l,i) / real(n_mc, wp)
              cr_dist(k,l,i) = cr_dist(k,l,i) / real(n_mc, wp)
           end do
        end do
     end if

     ! Stop timer
     call toc()
  end do

  call print_estimates()
  if (do_cr) then
     call print_confidence_regions()
  end if

  !----- Termination ----------------------------------------------------------!

  ! Free memory
  call feb_model_free(mod)
  deallocate(dat)
  deallocate(est)
  deallocate(est_mean)
  deallocate(est_cover)
  deallocate(est_dist)
  deallocate(est_sd)
  if (do_cr) then
     deallocate(dat_ss)
     deallocate(cr)
     deallocate(cr_mean)
     deallocate(cr_cover)
     deallocate(cr_dist)
  end if

contains

  !===== Internal Procedures ==================================================!

  subroutine run_experiment(dat, dat_ss, est, cr)
    type(dataset_t), intent(in) :: dat
    type(dataset_t), dimension(n_ss), intent(in) :: dat_ss
    real(wp), dimension(2, n_const), intent(out) :: est
    real(wp), dimension(2, n_alpha, n_const), intent(out) :: cr

    ! Local variables
    real(wp) :: tau                               ! Slackness
    real(wp), dimension(n_grid) :: Q              ! Objective function
    real(wp), dimension(n_ss) :: infQ             ! Subsampled estimates of infQ
    real(wp) :: c_hat                             ! (1-alpha) quantile of infQ
    integer :: n                                  ! Sample size
    integer :: j, k, l                            ! Generic indices

    ! Sample size
    n = dat%n

    ! Evaluate the sample objective function on the grid.
    Q = eval_sample_objective(dat, beta_grid)

    ! For each slackness sequence (each constant term)
    do l = 1, n_const
       ! Slackness
       tau = const(l) * sqrt(log(real(n, wp)))

       ! Estimate the identified set.
       call get_max_set(beta_grid, Q, n, tau, est(:,l))

       ! Determine infQ for each subsample using the estimated set.
       if (do_cr) then
          do j = 1, n_ss
             infQ(j) = get_infQ(dat_ss(j), est(:,l))
          end do
          call osl_quicksort(infQ)

          ! Construct confidence regions for several values of alpha.
          do k = 1, n_alpha
             ! Obtain the 1 - alpha quantile of the infQ values
             c_hat = quantile(infQ, 1.0_wp - alpha(k))

             ! Obtain the confidence region.
             call get_max_set(beta_grid, Q, n, c_hat, cr(:,k,l))
          end do
       end if
    end do
  end subroutine run_experiment

  ! Draw a random subsample of size M from the given dataset.
  subroutine subsample_dat(dat, m, dat_ss)
    type(dataset_t), intent(in) :: dat
    integer, intent(in) :: m
    type(dataset_t), intent(inout) :: dat_ss
    integer, dimension(m) :: idx

    ! obtain random subsample indices
    idx = random_k_subset(dat%n, dat_ss%n)

    ! populate the dat_ss dataset (assume it has been initialized)
    dat_ss%y = dat%y(:,idx)
    dat_ss%x = dat%x(:,:,idx)
  end subroutine subsample_dat

  ! Determine the value of the inferential statistic given a
  ! particular subsample.
  function get_infQ(dat_ss, est_set)
    type(dataset_t), intent(in) :: dat_ss
    real(wp), dimension(2), intent(in) :: est_set
    real(wp) :: get_infQ
    real(wp), dimension(n_grid) :: grid
    real(wp), dimension(n_grid) :: bmQ

    ! Construct a grid over the estimated set.x
    grid = osl_linspace(est_set(1), est_set(2), n_grid)

    ! Evaluate the objective function at the grid points.
    bmQ = eval_sample_objective(dat_ss, grid)

    ! Normalize and rescale.
    bmQ = sqrt(real(dat_ss%n, wp)) * (bmQ - maxval(bmQ))

    ! Return the negative of the minimum value of b_m * Q_m^(j).
    get_infQ = -minval(bmQ)
  end function get_infQ

  ! Evaluates the sample objective function with the given dataset
  ! over a fixed grid.
  function eval_sample_objective(dat, grid) result(Q)
    type(dataset_t), intent(in) :: dat
    real(wp), dimension(n_grid), intent(in) :: grid
    real(wp), dimension(n_grid) :: Q
    integer :: k

    do k = 1, n_grid
       call feb_model_sample_objective(mod, dat, grid(k), Q(k))
    end do
  end function eval_sample_objective

  ! Obtain a random subset of size K from N elements without
  ! replacement.
  function random_k_subset(n, k)
    integer, intent(in) :: n, k
    integer, dimension(k) :: random_k_subset
    logical, dimension(n) :: idx
    integer :: count, i

    ! obtain k random indices
    count = 0
    idx = .false.
    do
       i = osl_rng_uniform_int(rng, 1, n)
       if (.not. idx(i)) then
          idx(i) = .true.
          count = count + 1
       end if
       if (count == k) exit
    end do

    ! convert indices to combination
    count = 0
    do i = 1, n
       if (idx(i)) then
          count = count + 1
          random_k_subset(count) = i
       end if
    end do
  end function random_k_subset

  ! Prints a summary of parameters on startup.
  subroutine print_summary()
    print '(a)', '% Fixed Effects Model Monte Carlo Experiments'
    print '(a)', '% ==========================================='
    print '(a)', '%'
    print '(a)', '% Revision ' // trim(revision) // '.'
    print '(a)', '%'
    print '(a)', '% Population'
    print '(a)', '% ----------'
    print '(a)', '%'

    call print_vec('% x_1 ~ Uniform(', x1_range, 'f4.1', ', ', ')')
    call print_vec('% x_2 ~ Uniform({ ', x2_sup, 'f4.1', ', ', ' })')
    call print_vec('% \theta_0 = ( ', (/ beta0 /), 'f5.2', ', ', ' )')
    call print_vec('% \theta_I = [ ', id_set, 'f8.5', ', ', ' ]')

    print '(a)',    '%'
    print '(a)',    '% Monte Carlo Experiments'
    print '(a)',    '% -----------------------'
    print '(a)',    '%'
    print '(a,i0)', '% Number of trials: ', n_mc
    if (do_cr) then
       print '(a,i0)', '% Number of subsamples: ', n_ss
    end if

    call print_vec('% Sample sizes: { ', samp_sz, 'i0', ', ', ' }')
    call print_vec('% Slackness sequence constants: {', const, 'f5.2', ', ', ' }')
    call print_vec('% Grid interval: [ ', (/ grid_lb, grid_ub /), 'f5.2', ', ', ']')

    print '(a,i0)',        '% Number of grid points: ', n_grid
    print '(a,f7.4)',      '% Grid resolution: ', (grid_ub - grid_lb) / real(n_grid, wp)
    if (do_cr) then
       print '(a,i0,a,i0,a)', '% Subsample fraction: m = n^(', ss_frac_num, '/', ss_frac_denom, ')'
    end if

    ! Format population values for later
    write (population, '(a,f6.3,a,a,f6.3,a,f6.3,a)') &
         '$\theta_0 = ', beta0, '$, ', &
         '$\Theta_I = [', id_set(1), ', ', id_set(2), ']$.'

    print '(a)', '%'
    print '(a)', '% Running experiments...'
    print '(a)', '%'
  end subroutine print_summary

  subroutine print_experiment()
    if (do_cr) then
       print '("% n = ",i0,", m = ",i0)', n, m
    else
       print '("% n = ",i0)', n
    end if
  end subroutine print_experiment

  ! Print table summarizing estimates.
  subroutine print_estimates()
    print '(a)', '%'
    print '(a)', '% Estimates'
    print '(a)', '% ---------'
    print '(a)', ''
    print '(a)', '\begin{table}[p]'
    print '(a)', '  \centering'
    print '(a)', '  \begin{footnotesize}'
    print '(a)', '  \begin{tabular}{rrccrr}'
    print '(a)', '    \hline'
    print '(a)', '    \hline'
    print '(a)', '    $C$ & $n$   & Mean $\hat\Theta_n$ & St. Dev. & Coverage   & $d_H$ \\'
    do l = 1, n_const
       print '("    \hline")'
       print '("    \multirow{10}{*}{", f0.2, "}")', const(l)
       do i = 1, n_samp_sz
          write (*, '("    & ",i8," & ")', advance='no') samp_sz(i)
          write (*, '("[ ",f6.3,", ",f6.3," ]  & ")', advance='no') est_mean(:,l,i)
          write (*, '("[ ",f6.3,", ",f6.3," ] &  ")', advance='no') est_sd(:,l,i)
          write (*, '(f5.2,"     & ")', advance='no') est_cover(l,i)
          write (*, '(f5.2," \\")') est_dist(l,i)
       end do
    end do
    print '(a)', '    \hline'
    print '(a)', '  \end{tabular}'
    print '(a)', '  \end{footnotesize}'
    print '(a)', '  \caption{Fixed effects model estimates. ' // trim(population) // ' }'
    print '(a)', '  \label{tab:feb:estimates}'
    print '(a)', '\end{table}'
  end subroutine print_estimates

  ! Print table summarizing confidence regions.
  subroutine print_confidence_regions()
    print '(a)', '%'
    print '(a)', '% Confidence Regions'
    print '(a)', '% ------------------'
    print '(a)', ''
    print '(a)', '\begin{table}'
    print '(a)', '  \centering'
    print '(a)', '  \begin{footnotesize}'
    print '(a)', '  \begin{tabular}{rrrrrr}'
    print '(a)', '    \hline'
    print '(a)', '    \hline'
    print '(a)', '    & & & \multicolumn{4}{c}{Empirical Coverage} \\'
    print '(a)', '    $C$  & $m$  & $n$    &  0.750 &  0.900 &  0.950 \\'
    do l = 1, n_const
       print '(a)', '    \hline'
       print '("    \multirow{10}{*}{",f0.2,"} & \multirow{10}{*}{$n^{",i0,"/",i0,"}$}")', &
            const(l), ss_frac_num, ss_frac_denom
       do i = 1, n_samp_sz
          write (*, '("    & ")', advance='no')
          if (i == 1) then
             write (*, '("  ")', advance='no')
          else
             write (*, '("& ")', advance='no')
          end if
          write (*, '(i8," & ")', advance='no') samp_sz(i)
          write (*, '(f6.3," & ",f6.3," & ",f6.3," \\")') cr_cover(:,l,i)
       end do
    end do

    print '(a)', '    \hline'
    print '(a)', '  \end{tabular}'
    print '(a)', '  \end{footnotesize}'
    write (*, '("  \caption{Fixed effects model confidence regions, $m = n^{",i0,"/",i0,"}$.}")') ss_frac_num, ss_frac_denom
    write (*, '("  \label{tab:feb:conf:",f0.2,"}")') ss_frac
    print '(a)', '\end{table}'
  end subroutine print_confidence_regions

end program mcfeb
