! mcmph.f90 -- mixed proportional hazard Monte Carlo experiments
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program mcmph
  use osl_core, wp => osl_wp
  use osl_random
  use osl_stat, only: osl_sample_stats
  use osl_sort, only: osl_quicksort
  use dataset
  use mph_model
  use utility
  implicit none

  include 'revision.inc'

  !===== User Settings ========================================================!

  !----- Monte Carlo Parameters -----------------------------------------------!

  ! Number of trials for each sample size and slackness sequence
  integer, parameter :: n_mc = 1000

  ! Sample sizes
  integer, parameter :: n_samp_sz = 8
  integer, parameter, dimension(n_samp_sz) :: samp_sz &
       = (/ 125, 250, 500, 1000, 2000, 4000, 8000, 16000 /)

  ! Slackness sequences (proportional to sqrt(log(n)))
  integer, parameter :: n_const = 4
  real(wp), dimension(n_const), parameter :: const &
       = (/ 0.20_wp, 0.10_wp, 0.05_wp, 0.00_wp /)

  !----- Model Specification --------------------------------------------------!

  ! Population parameters
  real(wp), parameter :: alpha = 2.0_wp
  real(wp), parameter :: beta0 = 0.1_wp

  ! Regressor distribution (discrete and uniform)
  integer, parameter :: num_x1 = 3
  integer, parameter :: num_x2 = 5
  real(wp), dimension(num_x1), parameter :: x1 = (/ -1.0_wp, 0.0_wp, 1.0_wp /)
  real(wp), dimension(num_x2), parameter :: x2 = (/ -1.0_wp, -0.5_wp, -0.25_wp, 0.5_wp, 1.0_wp /)

  !----- Grid search options --------------------------------------------------!

  integer, parameter :: n_grid = 600                 ! Number of grid points
  real(wp) :: grid_lb = -1.5_wp                      ! Lower grid bounds
  real(wp) :: grid_ub =  1.5_wp                      ! Upper grid bounds

  !===== Internal Variables ===================================================!

  ! Evaluate Q and Q_N on a grid of size N_GRID.
  real(wp), dimension(n_grid) :: beta_grid           ! Grid support points
  real(wp), dimension(n_grid) :: Q_pop               ! Population obj. fn.

  ! Data
  type(osl_rng_t) :: rng                             ! Random number generator
  integer, parameter :: n_t = mph_model_n_t          ! Time dimension (fixed)
  integer :: n                                       ! Sample size
  type(dataset_t), dimension(:), allocatable :: dat  ! Datasets

  ! Population characteristics
  type(mph_model_t) :: mod                           ! Model
  real(wp), dimension(2) :: theta_tilde              ! \tilde\Theta
  character(len=100) :: population                   ! Reporting

  ! Results
  real(wp), dimension(:,:,:), allocatable :: est

  ! Statistics
  real(wp), dimension(:,:,:), allocatable :: est_mean
  real(wp), dimension(:,:), allocatable :: est_cover
  real(wp), dimension(:,:), allocatable :: est_dist
  real(wp), dimension(:,:,:), allocatable :: est_sd

  ! Indices
  integer :: i, r, l

  !===== Main Program =========================================================!

  !----- Preparations ---------------------------------------------------------!

  ! Allocate memory for storing results
  allocate(dat(n_mc))
  allocate(est(2, n_const, n_mc))
  allocate(est_mean(2, n_const, n_samp_sz))
  allocate(est_cover(n_const, n_samp_sz))
  allocate(est_dist(n_const, n_samp_sz))
  allocate(est_sd(2, n_const, n_samp_sz))

  ! Seed the PRNG
  call osl_rng_seed(rng, 2480129)

  ! Initialize the model
  call mph_model_init(mod, alpha, x1, x2)

  ! Construct a grid for beta
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)

  !----- Determine \tilde\Theta -----------------------------------------------!

  ! Obtain \tilde\Theta by evaluating population objective function on grid
  do i = 1, n_grid
     call mph_model_population_objective(mod, beta0, beta_grid(i), Q_pop(i))
  end do
  call get_max_set(beta_grid, Q_pop, 1, epsilon(1.0_wp), theta_tilde)

  !----- Print Summary --------------------------------------------------------!

  call print_summary()

  !----- Monte Carlo Experiments ----------------------------------------------!

  ! For each sample size
  do i = 1, n_samp_sz
     call tic()
     n = samp_sz(i)                               ! Sample size
     call print_experiment()

     ! Generate all data and subsamples in a critical region, using
     ! a common PRNG, so that the results will be reproducible when
     ! running in parallel.  Although the order in which datasets
     ! are generated and processed may differ, summary statistics
     ! will be unchanged.

     !$OMP PARALLEL DO
     do r = 1, n_mc
        !$OMP CRITICAL
        ! Allocate memory and generate data
        call dataset_init(dat(r), n, mph_model_n_t, mph_model_dim_x)
        call mph_model_generate_data(mod, beta0, rng, dat(r))
        !$OMP END CRITICAL

        ! Perform a single Monte Carlo trial and store the results.
        call run_experiment(dat(r), est(:,:,r))

        ! Free memory
        call dataset_free(dat(r))
     end do
     !$OMP END PARALLEL DO

     ! Average estimate, coverage, and mean Hausdorff distance.
     do l = 1, n_const
        call osl_sample_stats(est(:,l,:), MEAN=est_mean(:,l,i), SD=est_sd(:,l,i))
        est_cover(l,i) = 0.0_wp
        est_dist(l,i) = 0.0_wp
        do r = 1, n_mc
           if (cover_p(est(:,l,r), theta_tilde)) then
              est_cover(l,i) = est_cover(l,i) + 1.0_wp
           end if
           est_dist(l,i) = est_dist(l,i) + hausdorff(est(:,l,r), theta_tilde)
        end do
        est_cover(l,i) = est_cover(l,i) / real(n_mc, wp)
        est_dist(l,i) = est_dist(l,i) / real(n_mc, wp)
     end do

     ! Stop timer
     call toc()
  end do

  call print_results()

  !----- Termination ----------------------------------------------------------!

  ! Free memory for results
  deallocate(dat)
  deallocate(est)
  deallocate(est_mean)
  deallocate(est_cover)
  deallocate(est_dist)
  deallocate(est_sd)

  ! Free memory used for model
  call mph_model_free(mod)

contains

  !===== Internal Procedures ==================================================!

  subroutine run_experiment(dat, est)
    type(dataset_t), intent(in) :: dat
    real(wp), dimension(2, n_const), intent(out) :: est

    ! Local variables
    real(wp) :: tau                               ! Slackness
    real(wp), dimension(n_grid) :: Q              ! Objective function
    integer :: k, l                               ! Generic indices

    ! Evaluate the sample objective function on the grid.
    do k = 1, n_grid
       call mph_model_sample_objective(mod, dat, beta_grid(k), Q(k))
    end do

    ! Obtain an estimate for each slackness sequence
    do l = 1, n_const
       tau = const(l) * sqrt(log(real(dat%n, wp)))
       call get_max_set(beta_grid, Q, dat%n, tau, est(:,l))
    end do
  end subroutine run_experiment

  subroutine print_summary()
    print '(a)', '% Mixed Proportional Hazard Model Monte Carlo Experiments'
    print '(a)', '% ======================================================='
    print '(a)', '%'
    print '(a)', '% Revision ' // trim(revision) // '.'
    print '(a)', '%'
    print '(a)', '% Population'
    print '(a)', '% ----------'
    print '(a)', '%'

    call print_vec('% \theta_0 = ( ', (/ beta0 /), 'f5.2', ', ', ' )')
    call print_vec('% x_1 ~ Uniform({ ', x1, 'f4.1', ', ', ' })')
    call print_vec('% x_2 ~ Uniform({ ', x2, 'f4.1', ', ', ' })')
    call print_vec('% \tilde\Theta = [ ', theta_tilde, 'f6.3', ', ', ' ]')

    print '(a)',    '%'
    print '(a)',    '% Monte Carlo Experiments'
    print '(a)',    '% -----------------------'
    print '(a)',    '%'
    print '(a,i0)', '% Number of trials: ', n_mc

    call print_vec('% Sample sizes: { ', samp_sz, 'i0', ', ', ' }')
    call print_vec('% Slackness sequence constants: {', const, 'f5.2', ', ', ' }')
    call print_vec('% Grid interval: [ ', (/ grid_lb, grid_ub /), 'f5.2', ', ', ']')

    print '(a,i0)',        '% Number of grid points: ', n_grid
    print '(a,f7.4)',      '% Grid resolution: ', (grid_ub - grid_lb) / real(n_grid, wp)

    ! Format population values for later
    write (population, '(a,f6.3,a,a,f6.3,a,f6.3,a)') &
         '$\theta_0 = (', beta0, ')$, ', &
         '$\tilde\Theta = [', theta_tilde(1), ', ', theta_tilde(2), ']$.'

    print '(a)', '%'
    print '(a)', '% Running experiments...'
    print '(a)', '%'
  end subroutine print_summary

  subroutine print_experiment()
    print '("% n = ",i0)', n
  end subroutine print_experiment

  ! Prints tables with summary statistics.
  subroutine print_results()
    print '(a)', '%'
    print '(a)', '% Estimates'
    print '(a)', '% ---------'
    print '(a)', ''
    print '(a)', '\begin{table}[p]'
    print '(a)', '  \centering'
    print '(a)', '  \begin{footnotesize}'
    print '(a)', '  \begin{tabular}{rrccrr}'
    print '(a)', '    \hline'
    print '(a)', '    \hline'
    print '(a)', '    $\kappa_n$ & $n$   & Mean $\hat\Theta_n$ & St. Dev.  & Coverage & $d_H$ \\'
    do l = 1, n_const
       print '("    \hline")'
       print '("    \multirow{10}{*}{$", f0.2, "\sqrt{\ln n}$}")', const(l)
       do i = 1, n_samp_sz
          write (*, '("    & ",i8," & ")', advance='no') samp_sz(i)
          write (*, '("[ ",f6.3,", ",f6.3," ]  & ")', advance='no') est_mean(:,l,i)
          write (*, '("[ ",f6.3,", ",f6.3," ] &  ")', advance='no') est_sd(:,l,i)
          write (*, '(f5.2,"     & ")', advance='no') est_cover(l,i)
          write (*, '(f5.2," \\")') est_dist(l,i)
       end do
    end do
    print '(a)', '    \hline'
    print '(a)', '  \end{tabular}'
    print '(a)', '  \end{footnotesize}'
    print '(a)', '  \caption{Mixed proportional hazard model estimates. ' // trim(population) // ' }'
    print '(a)', '  \label{tab:mph:estimates}'
    print '(a)', '\end{table}'
  end subroutine print_results

end program mcmph
