! mph_model.f90 -- mixed proportional hazard model
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

! This module implements a basic fixed effects mixed proportional
! hazard model.  The following assumptions are made:
!
! * There are T = 2 time periods.
! * There are two covariates with the coefficient on the first normalized to 1.
! * The distribution of X is discrete and uniform.
! * The unobservables u_t are iid with minus type I extreme value distributions.
! * The basline hazard function is Weibull with parameter alpha.
! * The fixed effect is the average of the components of x's across time.
module mph_model
  use osl_core, wp => osl_wp
  use osl_dist
  use osl_random
  use dataset
  implicit none

  private

  public :: mph_model_dim_x
  public :: mph_model_n_t

  public :: mph_model_t
  public :: mph_model_init
  public :: mph_model_free
  public :: mph_model_lambda
  public :: mph_model_generate_data
  public :: mph_model_population_objective
  public :: mph_model_sample_objective
  public :: mph_model_sample_obj_lam

  integer, parameter :: mph_model_dim_x = 2       ! Dimension of X_t
  integer, parameter :: mph_model_n_t = 2         ! Time periods

  type :: mph_model_t
     real(wp) :: alpha                            ! Weibull hazard parameter
     integer :: n_x                               ! Number of regressor values
     real(wp), dimension(:,:), pointer :: x       ! Joint regressor values
     type(osl_discrete_rnd_t) :: dist_x           ! Discrete X distribution
  end type mph_model_t

contains

  ! Initialize a new model and allocate memory.
  subroutine mph_model_init(self, alpha, x1, x2)
    type(mph_model_t), intent(inout) :: self
    real(wp), intent(in) :: alpha
    real(wp), dimension(:), intent(in) :: x1
    real(wp), dimension(:), intent(in) :: x2

    real(wp), dimension(:), allocatable :: px
    integer :: i, j, k, n_x1, n_x2

    ! Store parameters
    self%alpha = alpha

    ! Determine cardinality of support of X and allocate memory
    n_x1 = size(x1)
    n_x2 = size(x2)
    self%n_x = n_x1 * n_x2
    allocate(self%x(2,self%n_x))
    allocate(px(self%n_x))

    ! Construct support of x
    i = 0
    do j = 1, n_x1
       do k = 1, n_x2
          i = i + 1
          self%x(1,i) = x1(j)
          self%x(2,i) = x2(k)
       end do
    end do

    ! Construct point masses for uniform X
    px = 1.0_wp / real(self%n_x, wp)

    ! Prepare to the discrete distribution of regressors
    call osl_discrete_rnd_init(self%dist_x, px)

    ! Clean up
    deallocate(px)
  end subroutine mph_model_init

  ! Destroy an existing model, freeing any allocated memory.
  subroutine mph_model_free(self)
    type(mph_model_t), intent(inout) :: self
    
    ! Free the discrete distribution
    call osl_discrete_rnd_free(self%dist_x)

    ! Free dynamically allocated memory
    deallocate(self%x)
  end subroutine mph_model_free

  ! Evaluate the transformation function.
  function mph_model_lambda(self, y) result(lambda)
    type(mph_model_t), intent(in) :: self
    real(wp), intent(in) :: y
    real(wp) :: lambda

    lambda = self%alpha * log(y)
  end function mph_model_lambda

  ! Evaluate the inverse transformation function.
  function mph_model_lambda_inverse(self, z) result(y)
    type(mph_model_t), intent(in) :: self
    real(wp), intent(in) :: z
    real(wp) :: y

    y = exp(z / self%alpha)
  end function mph_model_lambda_inverse

  ! Return the value of the individual effect given x_1 and x_2.
  function mph_model_generate_c(self, x1, x2) result(c)
    type(mph_model_t), intent(in) :: self
    real(wp), dimension(2), intent(in) :: x1
    real(wp), dimension(2), intent(in) :: x2
    real(wp) :: c

    c = sum(x1 + x2) / real(mph_model_n_t * mph_model_dim_x, wp)
  end function mph_model_generate_c

  ! Generate a sample from the model.
  subroutine mph_model_generate_data(self, beta, rng, dat)
    type(mph_model_t), intent(inout) :: self
    real(wp), intent(in) :: beta
    type(osl_rng_t), intent(inout) :: rng
    type(dataset_t), intent(inout) :: dat

    integer :: ii, it, ix
    real(wp) :: xbeta, c, index
    real(wp), dimension(dat%t) :: u

    ! Sanity checks
    if (dat%t /= mph_model_n_t) then
       stop 'The MPH model is only implemented for T = 2.'
    end if

    ! Generate data for each individual ii
    do ii = 1, dat%n
       ! Generate regressors and disturbances for each period it
       do it = 1, dat%t
          ! Regressors
          ix = osl_discrete_rnd(self%dist_x, rng)
          dat%x(:,it,ii) = self%x(:,ix)

          ! Disturbances: minus extreme value distribution
          u(it) = -1.0_wp * osl_std_gumbel_rnd(rng)
       end do

       ! Generate the fixed effect for individual ii
       c = mph_model_generate_c(self, dat%x(:,1,ii), dat%x(:,2,ii))

       ! Determine the duration of each spell
       do it = 1, dat%t
          xbeta = dat%x(1,it,ii) + beta * dat%x(2,it,ii)
          index = xbeta + c + u(it)
          dat%y(it,ii) = mph_model_lambda_inverse(self, index)
       end do
    end do
  end subroutine mph_model_generate_data

  ! Evaluate the population objective function at theta
  subroutine mph_model_population_objective(self, beta0, beta, Q)
    type(mph_model_t), intent(in) :: self
    real(wp), intent(in) :: beta0
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q

    real(wp) :: dx1, dx2, dxbeta0, dxbeta, pr
    integer :: i, j

    Q = 0.0_wp
    do i = 1, self%n_x
       do j = 1, self%n_x
          ! Difference regressors over time
          dx1 = self%x(1,i) - self%x(1,j)
          dx2 = self%x(2,i) - self%x(2,j)

          ! Evaluate the index and population probabilities (beta0)
          dxbeta0 = dx1 + beta0 * dx2
          pr = exp(dxbeta0)
          pr = 2.0_wp * pr / (1.0_wp + pr) - 1.0_wp

          ! Evaluate the index using beta
          dxbeta = dx1 + beta * dx2

          ! Accumulate the value for this (x1, x2) pair
          Q = Q + pr * osl_sgn(dxbeta)
       end do
    end do
    Q = Q / real(self%n_x * self%n_x, wp)
  end subroutine mph_model_population_objective

  ! Evaluate the sample objective function at theta
  subroutine mph_model_sample_objective(self, dat, beta, Q)
    type(mph_model_t), intent(in) :: self
    type(dataset_t), intent(in) :: dat
    real(wp), intent(in) :: beta
    real(wp), intent(out) :: Q
    real(wp) :: dx1, dx2
    integer :: i

    ! Calculate the sample average of sgn(y2 - y1) * sgn((x2-x1)'beta).
    Q = 0.0_wp
    do i = 1, dat%n
       dx1 = dat%x(1,2,i) - dat%x(1,1,i)
       dx2 = dat%x(2,2,i) - dat%x(2,1,i)
       Q = Q + osl_sgn(dat%y(2,i) - dat%y(1,i)) * osl_sgn(dx1 + beta * dx2)
    end do
    Q = Q / real(dat%n, wp)
  end subroutine mph_model_sample_objective

  ! Evaluate the sample objective function which characterizes \Lambda(y)
  ! for a given value of y, given an estimate of beta.
  subroutine mph_model_sample_obj_lam(self, dat, beta, y_bar, y, lambda, Q)
    type(mph_model_t), intent(in) :: self
    type(dataset_t), intent(in) :: dat
    real(wp), intent(in) :: beta
    real(wp), intent(in) :: y_bar
    real(wp), intent(in) :: y
    real(wp), intent(in) :: lambda
    real(wp), intent(out) :: Q

    real(wp) :: dx1, dx2, dxbeta, y1, y2
    integer :: i

    Q = 0.0_wp
    do i = 1, dat%n
       y1 = dat%y(1,i)
       y2 = dat%y(2,i)
       dx1 = dat%x(1,2,i) - dat%x(1,1,i)
       dx2 = dat%x(2,2,i) - dat%x(2,1,i)
       dxbeta = dx1 + beta * dx2

       ! Accumulate objective function contribution of this observation
       Q = Q + (ind(y2 > y) - ind(y1 > y_bar)) * ind(dxbeta > lambda)
       Q = Q + (ind(y2 > y_bar) - ind(y1 > y)) * ind(dxbeta > -lambda)
    end do
    Q = Q / real(2 * dat%n, wp)

    contains

      function ind(expr)
        logical, intent(in) :: expr
        real(wp) :: ind

        if (expr) then
           ind = 1.0_wp
        else
           ind = 0.0_wp
        end if
      end function ind
  end subroutine mph_model_sample_obj_lam

end module mph_model
