! osl_sort.f90 -- sorting routines
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_sort
  use osl_core, wp => osl_wp
  implicit none

  private

  public :: osl_quicksort
  public :: osl_shellsort

  interface osl_quicksort
     module procedure osl_quicksort_vwp
     module procedure osl_quicksort_vi
     module procedure osl_quicksort_generic
  end interface osl_quicksort

contains

  ! Sort an array of real numbers in ascending order.  This is an
  ! implementation of the Quicksort algorithm of C.A.R. Hoare (1961).
  !
  ! *References*
  !
  !  - Hoare, C. A. R. (1961). Quicksort: Algorithm 64.
  !    _Communications of the ACM_ 4, 321-322.
  subroutine osl_quicksort_vwp(v)
    real(wp), dimension(:), intent(inout) :: v
    integer, parameter :: m = 7
    integer :: n_stack = 50
    integer, dimension(:), allocatable :: s, s_temp
    integer :: i, ir, j, k, jstack, l, n
    real(wp) :: a

    ! initialize stack
    allocate(s(n_stack))

    n = size(v)

    jstack = 0
    l = 1
    ir = n

    do
       if (ir - l < m) then
          do j = l + 1, ir
             a = v(j)
             do i = j - 1 , 1, -1
                if (v(i) <= a) exit
                v(i+1) = v(i)
             end do
             v(i+1) = a
          end do

          if (jstack == 0) return

          ir = s(jstack)
          l = s(jstack-1)
          jstack = jstack - 2

       else

          k = (l+ir)/2
          call osl_swap(v(k),v(l+1))
          if (v(l+1) > v(ir)) call osl_swap(v(l+1), v(ir))
          if (v(l) > v(ir)) call osl_swap(v(l), v(ir))
          if (v(l+1) > v(l)) call osl_swap(v(l+1), v(l))

          i = l + 1
          j = ir
          a = v(l)

          do
             do
                i = i + 1
                if (v(i) >= a) exit
             end do

             do
                j = j - 1
                if (v(j) <= a) exit
             end do

             if (j < i) exit
             call osl_swap(v(i), v(j))
          end do

          v(l) = v(j)
          v(j) = a
          jstack = jstack + 2

          ! double the stack size if necessary
          if (jstack > n_stack) then
             allocate(s_temp(n_stack))
             s_temp = s
             deallocate(s)
             n_stack = 2 * n_stack
             allocate(s(n_stack))
             s(1:size(s_temp,1)) = s_temp
             deallocate(s_temp)
          end if

          if (ir - i + 1 > j - l) then
             s(jstack) = ir
             s(jstack-1) = i
             ir = j-1
          else
             s(jstack) = j-1
             s(jstack-1) = l
             l = i
          end if
       end if
    end do

    ! free stack
    deallocate(s)
  end subroutine osl_quicksort_vwp

  ! Sort an array of integers in ascending order.  This is an
  ! implementation of the Quicksort algorithm of C.A.R. Hoare (1961).
  !
  ! *References*
  !
  !  - Hoare, C. A. R. (1961). Quicksort: Algorithm 64.
  !    _Communications of the ACM_ 4, 321-322.
  subroutine osl_quicksort_vi(v)
    integer, dimension(:), intent(inout) :: v
    integer, parameter :: m = 7
    integer :: n_stack = 50
    integer, dimension(:), allocatable :: s, s_temp
    integer :: i, ir, j, k, jstack, l, n
    integer :: a

    ! initialize stack
    allocate(s(n_stack))

    n = size(v)

    jstack = 0
    l = 1
    ir = n

    do
       if (ir - l < m) then
          do j = l + 1, ir
             a = v(j)
             do i = j - 1 , 1, -1
                if (v(i) <= a) exit
                v(i+1) = v(i)
             end do
             v(i+1) = a
          end do

          if (jstack == 0) return

          ir = s(jstack)
          l = s(jstack-1)
          jstack = jstack - 2

       else

          k = (l+ir)/2
          call osl_swap(v(k),v(l+1))
          if (v(l+1) > v(ir)) call osl_swap(v(l+1), v(ir))
          if (v(l) > v(ir)) call osl_swap(v(l), v(ir))
          if (v(l+1) > v(l)) call osl_swap(v(l+1), v(l))

          i = l + 1
          j = ir
          a = v(l)

          do
             do
                i = i + 1
                if (v(i) >= a) exit
             end do

             do
                j = j - 1
                if (v(j) <= a) exit
             end do

             if (j < i) exit
             call osl_swap(v(i), v(j))
          end do

          v(l) = v(j)
          v(j) = a
          jstack = jstack + 2

          ! double the stack size if necessary
          if (jstack > n_stack) then
             allocate(s_temp(n_stack))
             s_temp = s
             deallocate(s)
             n_stack = 2 * n_stack
             allocate(s(n_stack))
             s(1:size(s_temp,1)) = s_temp
             deallocate(s_temp)
          end if

          if (ir - i + 1 > j - l) then
             s(jstack) = ir
             s(jstack-1) = i
             ir = j-1
          else
             s(jstack) = j-1
             s(jstack-1) = l
             l = i
          end if
       end if
    end do

    ! free stack
    deallocate(s)
  end subroutine osl_quicksort_vi

  ! Sort a generic collection of N objects using the quicksort
  ! algorithm with the generic COMPARE and SWAP routines and the
  ! provided context CTX.
  subroutine osl_quicksort_generic(n, compare, swap, ctx)
    integer, intent(in) :: n
    integer, dimension(:), intent(in) :: ctx

    interface

       function compare(i, j, ctx)
         integer, intent(in) :: i, j
         integer, dimension(:), intent(in), optional :: ctx
         integer :: compare
       end function compare

       subroutine swap(i, j, ctx)
         integer, intent(in) :: i, j
         integer, dimension(:), intent(in), optional :: ctx
       end subroutine swap

    end interface

    integer, parameter :: small = 6

    if (n == 2) then
       if (compare(2, 1, ctx) == 1) then
          call swap(2, 1, ctx);
       end if
    else
       call qsort(1, n)
    end if

  contains

      recursive subroutine qsort(l, r)
        integer, intent(in) :: l, r
        integer :: i, j, mid

        if (r < l + small) then
           call interchange_sort(l, r)
        else
           mid = (l +  r) / 2
           i = l - 1
           j = r + 1

           do
              ! Search from left for an element larger than the one at mid
              do
                 i = i + 1
                 if (compare(i, mid, ctx) == -1) exit
              end do
              ! Search from the right for an element weakly smaller than mid
              do
                 j = j - 1
                 if (compare(mid, j, ctx) == -1) exit
              end do

              if (i < j) then
                 ! Elements I and J are out of order, swap them
                 call swap(i, j, ctx)
              else if (i == j) then
                 i = i + 1
                 exit
              else
                 exit
              end if
           end do

           if (l < j) call qsort(l, j)
           if (i < r) call qsort(i, r)
        end if
      end subroutine qsort

      subroutine interchange_sort(l, r)
        integer, intent(in) :: l, r
        integer :: i, j

        do i = l, r - 1
           do j = i + 1, r
              if (compare(j, i, ctx) == 1) then
                 call swap(i, j, ctx)
              end if
           end do
        end do
      end subroutine interchange_sort

  end subroutine osl_quicksort_generic

  ! Sort an array V of real numbers in ascending order using the
  ! algorithm of Shell (1959).
  !
  ! *References*
  !
  !  - Shell, D. L. (1959).  A high-speed sorting procedure.
  !    _Communications of the ACM_ 2, 30-32.
  subroutine osl_shellsort(v)
    real(wp), dimension(:), intent(inout) :: v
    integer :: inc, i, j
    real(wp) :: temp

    inc = nint(size(v) / 2.0_wp)
    do while (inc > 0)
       do i = inc, size(v)
          temp = v(i)
          j = i
          do while (j > inc)
             if (v(j - inc) <= temp) then
                exit
             end if
             v(j) = v(j - inc)
             j = j - inc
          end do
          v(j) = temp
       end do
       inc = nint(inc / 2.2_wp)
    end do
  end subroutine osl_shellsort

end module osl_sort
