! plotlam.f90 -- transformation function plots for the MPH model
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program plotlam
  use osl_core, wp => osl_wp
  use osl_random
  use dataset
  use mph_model
  use utility
  implicit none

  !===== User-Defined Settings ===============================================!

  !----- Plotting Options ----------------------------------------------------!

  ! Set OPT_XFIG to .TRUE. to generate an xfig .fig file
  logical :: opt_xfig = .false.

  !----- Model Specification -------------------------------------------------!

  ! Population parameters
  real(wp), parameter :: alpha0 = 2.0_wp
  real(wp), parameter :: beta0 = 0.1_wp

  ! Regressor distribution (discrete and uniform)
  integer, parameter :: num_x1 = 3
  integer, parameter :: num_x2 = 5
  real(wp), dimension(num_x1), parameter :: x1 = (/ -1.0_wp, 0.0_wp, 1.0_wp /)
  real(wp), dimension(num_x2), parameter :: x2 = (/ -1.0_wp, -0.5_wp, -0.25_wp, 0.5_wp, 1.0_wp /)

  !----- Sample Specification ------------------------------------------------!

  integer, parameter :: nn = 250

  !----- Estimator -----------------------------------------------------------!

  real(wp), parameter :: const = 0.05_wp          ! Slackness sequence constant

  !----- Grid Search Options -------------------------------------------------!

  integer, parameter :: n_grid = 100              ! Number of grid points
  real(wp), parameter :: beta_lb = -3.0_wp        ! Lower bound for beta
  real(wp), parameter :: beta_ub =  3.0_wp        ! Upper bound for beta
  real(wp), parameter :: y_lb =  0.25_wp          ! Lower bound for y
  real(wp), parameter :: y_ub =  2.25_wp          ! Upper bound for y
  real(wp), parameter :: lam_lb = -5.0_wp         ! Lower bound for lambda
  real(wp), parameter :: lam_ub =  5.0_wp         ! Upper bound for lambda

  !===== Internal Variables ==================================================!

  integer, parameter :: lun = 25                  ! Output unit number
  integer, parameter :: n_t = mph_model_n_t       ! Time dimension (fixed)
  real(wp), dimension(n_grid) :: beta_grid        ! Grid for beta
  real(wp), dimension(n_grid) :: y_grid           ! Grid for y
  real(wp), dimension(n_grid) :: lam_grid         ! Grid for Lambda
  real(wp), dimension(n_grid) :: Q                ! Population objective fn.
  real(wp), dimension(n_grid) :: Qn, Qn1, Qn2     ! Sample objective fn.
  real(wp), dimension(2) :: beta_pop              ! Population bounds for beta
  real(wp), dimension(2) :: beta_est              ! Estimated bounds for beta
  real(wp), dimension(n_grid) :: lam_true         ! True values of Lambda
  real(wp), dimension(n_grid) :: lam_est_lb       ! Lower estimate of Lambda
  real(wp), dimension(n_grid) :: lam_est_ub       ! Upper estimate of Lambda
  real(wp), dimension(2) :: lam_est_1             ! Estimated set for Lambda
  real(wp), dimension(2) :: lam_est_2             ! Estimated set for Lambda
  real(wp) :: y_min, y_max, y_mean                ! Sample moments of y
  real(wp) :: lam_base                            ! Normalizing value
  type(osl_rng_t) :: rng                          ! Random number generator
  type(mph_model_t) :: mod                        ! Model
  type(dataset_t) :: dat                          ! Dataset storage
  real(wp) :: tau                                 ! Slackness
  integer :: j, k                                 ! Grid point index

  !===== Main Program ========================================================!

  ! Seed the PRNG
  call osl_rng_seed(rng, 9087423)

  ! Initialize the model
  call mph_model_init(mod, alpha0, x1, x2)

  ! Generate data
  call dataset_init(dat, nn, mph_model_n_t, mph_model_dim_x)
  call mph_model_generate_data(mod, beta0, rng, dat)

  ! Calculate sample statistics for y
  y_min = minval(dat%y)
  y_max = maxval(dat%y)
  y_mean = sum(dat%y) / real(n_t * nn, wp)
  print '(a,f7.4)', '# y_min =  ', y_min
  print '(a,f7.4)', '# y_max =  ', y_max
  print '(a,f7.4)', '# y_mean = ', y_mean

  ! Determine slackness
  tau = const * sqrt(log(real(nn, wp)))

  ! Open output file
  open(UNIT=lun, FILE='lambda.dat')

  ! Initialize grids
  beta_grid = osl_linspace(beta_lb, beta_ub, n_grid)
  y_grid = osl_linspace(y_lb, y_ub, n_grid)
  lam_grid = osl_linspace(lam_lb, lam_ub, n_grid)

  ! Population bounds for beta
  do k = 1, n_grid
     call mph_model_population_objective(mod, beta0, beta_grid(k), Q(k))
  end do
  call get_max_set(beta_grid, Q, 1, epsilon(1.0_wp), beta_pop)
  print '(a,2(f6.3,1x),a)', '# beta_pop = [ ', beta_pop, ']'

  ! Estimated bounds for beta
  do k = 1, n_grid
     call mph_model_sample_objective(mod, dat, beta_grid(k), Qn(k))
  end do
  call get_max_set(beta_grid, Qn, dat%n, tau, beta_est)
  print '(a,2(f6.3,1x),a)', '# beta_est = [ ', beta_est, ']'

  lam_base = mph_model_lambda(mod, y_mean)

  do k = 1, n_grid
     ! True value of Lambda(y)
     lam_true(k) = mph_model_lambda(mod, y_grid(k)) - lam_base

     ! Estimate bounds for Lambda(y)
     do j = 1, n_grid
        call mph_model_sample_obj_lam(mod, dat, beta_est(1), y_mean, y_grid(k), lam_grid(j), Qn1(j))
        call mph_model_sample_obj_lam(mod, dat, beta_est(2), y_mean, y_grid(k), lam_grid(j), Qn2(j))
     end do
     call get_max_set(lam_grid, Qn1, nn, tau, lam_est_1)
     call get_max_set(lam_grid, Qn2, nn, tau, lam_est_2)
     lam_est_lb(k) = min(minval(lam_est_1), minval(lam_est_2))
     lam_est_ub(k) = max(maxval(lam_est_1), maxval(lam_est_2))

     ! Write results to file
     write(lun, '(4g17.5)') y_grid(k), lam_true(k), lam_est_lb(k), lam_est_ub(k)
  end do

  ! Close output file
  close(lun)

  ! Print gnuplot commands
  if (opt_xfig) then
     print '(a)', 'set terminal fig textspecial'
     print '(a)', 'set output "lambda.fig"'
  end if
  print '(a,f5.2,a,f5.2,a)', 'set xrange [', y_lb, ':', y_ub, ']'
  print '(a,f5.2,a,f5.2,a,f5.2)', 'set xtic ', y_lb, ', ', (y_ub - y_lb) / 4.0_wp, ', ', y_ub
  print '(a)', 'set key right bottom'
  print '(a)', 'set size 0.8,0.8'
  print '(a)', 'set xlabel "$y$"'
  print '(a)', 'plot' &
       // ' "lambda.dat" using 1:3 lt 1 with lines title "$\\hat\\Lambda(y)$",' &
       // ' "lambda.dat" using 1:4 lt 1 with lines notitle,' &
       // ' "lambda.dat" using 1:2 lt 3 with lines title "$\\Lambda(y)$"'

  ! Free used memory
  call dataset_free(dat)
  call mph_model_free(mod)
end program plotlam
