! plotqfe.f90 -- plots the fixed effects model objective function
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program plotqfe
  use osl_core, wp => osl_wp
  use osl_random
  use dataset
  use fe_model
  implicit none

  !===== User-Defined Settings ===============================================!

  !----- Plotting Options ----------------------------------------------------!

  ! Set OPT_XFIG to .TRUE. to generate an xfig figure
  logical :: opt_xfig = .false.

  !----- Model Specification -------------------------------------------------!

  ! Population parameters
  real(wp), parameter :: beta0 = -0.15_wp

  ! Regressor distribution (discrete and uniform)
  integer, parameter :: num_x1 = 3
  integer, parameter :: num_x2 = 5
  real(wp), dimension(num_x1), parameter :: x1 = (/ -1, 0, 1 /)
  real(wp), dimension(num_x2), parameter :: x2 = (/ 0, 1, 2, 3, 4 /)

  !----- Sample Specification ------------------------------------------------!

  integer, parameter :: nn = 500                  ! Number of observations
  integer, parameter :: nt = 2                    ! Time dimension
  integer, parameter :: dim_x = 2                 ! Regressor dimension

  !----- Grid Search Options -------------------------------------------------!

  integer, parameter :: n_grid = 300
  real(wp) :: grid_lb = -3.0_wp
  real(wp) :: grid_ub = 3.0_wp

  type(osl_rng_t) :: rng
  type(fe_model_t) :: fe_mod
  type(dataset_t) :: dat

  !===== Internal Variables ==================================================!

  real(wp), dimension(n_grid) :: beta_grid        ! Grid support
  real(wp), dimension(n_grid) :: Q                ! Population objective fn.
  real(wp), dimension(n_grid) :: Q_n              ! Sample objective fn.
  integer :: k                                    ! Grid point index

  !===== Main Program ========================================================!

  ! Construct a grid for theta
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)

  ! Initialize the model
  call fe_model_init(fe_mod, x1, x2, beta0)

  ! Generate data
  call dataset_init(dat, nn, nt, dim_x)
  call fe_model_generate_data(fe_mod, rng, dat)

  ! Evaluate the population and sample objective functions
  do k = 1, n_grid
     call fe_model_population_objective(fe_mod, beta_grid(k), Q(k))
     call fe_model_sample_objective(fe_mod, dat, beta_grid(k), Q_n(k))
  end do

  ! Write data file
  open(UNIT=25, FILE='qfe.dat')
  do k = 1, n_grid
     write(25, '(3g17.5)') beta_grid(k), Q(k), Q_n(k)
  end do
  close(25)

  ! Generate plots
  if (opt_xfig) then
     print '(a)', 'set terminal fig textspecial'
     print '(a)', 'set output "qfe.fig"'
  end if
  print '(a)', 'set size 0.8,0.8'
  print '(a)', 'set xlabel "$\\beta$"'
  print '(a)', 'plot' &
       // ' "qfe.dat" using 1:2 with steps lt 1 title "$Q(\\beta)$",' &
       // ' "qfe.dat" using 1:3 with steps lt 3 title "$Q_n(\\beta)$"'

  ! Clean up
  call dataset_free(dat)
  call fe_model_free(fe_mod)
end program plotqfe
