! plotqfeb.f90 -- plots the fixed effects model objective function
!
! Copyright (C) 2009-2011 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program plotqfeb
  use osl_core, wp => osl_wp
  use osl_random
  use dataset
  use feb_model
  implicit none

  !===== User-Defined Settings ===============================================!

  !----- Plotting Options ----------------------------------------------------!

  ! Set OUTPUT to either 'fig', 'eps', or 'pdf' to produce a graphics file
  character(len=3), parameter :: output = 'pdf'

  !----- Model Specification -------------------------------------------------!

  ! Bounds on support of X_{1t}
  real(wp), parameter, dimension(2) :: x1_range = (/ 0.0_wp, 1.0_wp /)

  ! Finite support of X_{2t}
  real(wp), parameter, dimension(3) :: x2_sup = (/ 1.0_wp, 2.0_wp, 3.0_wp /)

  ! Population parameters
  real(wp), parameter :: beta0 = -0.6_wp

  ! Simulation
  integer, parameter :: nsim = 10000

  !----- Sample Specification ------------------------------------------------!

  integer, parameter :: nn = 100                  ! Number of observations

  !----- Grid Search Options -------------------------------------------------!

  integer, parameter :: n_grid = 6001
  real(wp) :: grid_lb = -3.0_wp
  real(wp) :: grid_ub = 3.0_wp

  type(osl_rng_t) :: rng
  type(feb_model_t) :: feb_mod
  type(dataset_t) :: dat

  !===== Internal Variables ==================================================!

  real(wp), dimension(n_grid) :: beta_grid        ! Grid support
  real(wp), dimension(n_grid) :: Q                ! Population objective fn.
  real(wp), dimension(n_grid) :: Q_n              ! Sample objective fn.
  integer :: k                                    ! Grid point index

  !===== Main Program ========================================================!

  ! Construct a grid for theta
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)

  ! Initialize the model
  call feb_model_init(feb_mod, x1_range, x2_sup, beta0, nsim)

  ! Generate data
  call osl_rng_seed(rng, 759827334)
  call dataset_init(dat, nn, n_t, dim_x)
  call feb_model_generate_data(feb_mod, rng, dat)

  ! Evaluate the population and sample objective functions
  do k = 1, n_grid
     call feb_model_population_objective(feb_mod, beta_grid(k), Q(k))
     call feb_model_sample_objective(feb_mod, dat, beta_grid(k), Q_n(k))
  end do

  ! Write data file
  open(UNIT=25, FILE='qfeb.dat')
  do k = 1, n_grid
     write(25, '(3g17.5)') beta_grid(k), Q(k), Q_n(k)
  end do
  close(25)

  ! Generate plots
  select case (output)
  case ('fig')
     print '(a)', 'set terminal fig textspecial'
     print '(a)', 'set output "qfeb.fig"'
  case ('eps')
     print '(a)', 'set terminal postscript eps enhanced'
     print '(a)', 'set output "qfeb.eps"'
  case ('pdf')
     print '(a)', 'set terminal pdf'
     print '(a)', 'set output "qfeb.pdf"'
  end select
  print '(a)', 'set size 0.8,0.8'
  print '(a)', 'set xlabel "$\\beta$"'
  print '(a)', 'plot' &
       // ' "qfeb.dat" using 1:2 with lines lt 1 title "$Q(\\beta)$",' &
       // ' "qfeb.dat" using 1:3 with steps lt 3 title "$Q_n(\\beta)$"'

  ! Clean up
  call dataset_free(dat)
  call feb_model_free(feb_mod)
end program plotqfeb
