! plotmph.f90 -- objective function plots for the MPH model
!
! Copyright (C) 2009-2010 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program plotqmph
  use osl_core, wp => osl_wp
  use osl_random
  use dataset
  use mph_model
  implicit none

  !===== User-Defined Settings ===============================================!

  !----- Plotting Options ----------------------------------------------------!

  ! Set OPT_XFIG to .TRUE. to generate an xfig .fig file
  logical :: opt_xfig = .false.

  !----- Model Specification -------------------------------------------------!

  ! Population parameters
  real(wp), parameter :: alpha = 2.0_wp
  real(wp), parameter :: beta0 = 0.1_wp

  ! Regressor distribution (discrete and uniform)
  integer, parameter :: num_x1 = 3
  integer, parameter :: num_x2 = 5
  real(wp), dimension(num_x1), parameter :: x1 = (/ -1.0_wp, 0.0_wp, 1.0_wp /)
  real(wp), dimension(num_x2), parameter :: x2 = (/ -1.0_wp, -0.5_wp, -0.25_wp, 0.5_wp, 1.0_wp /)

  !----- Sample Specification ------------------------------------------------!

  integer, parameter :: nn = 250

  !----- Grid Search Options -------------------------------------------------!

  integer, parameter :: grid_size = 3000          ! Number of grid points
  real(wp), parameter :: grid_lb = -3.0_wp        ! Lower bound
  real(wp), parameter :: grid_ub = 3.0_wp         ! Upper bound

  !===== Internal Variables ==================================================!

  integer, parameter :: lun = 25                  ! Output unit number
  integer, parameter :: n_t = mph_model_n_t       ! Time dimension (fixed)
  type(osl_rng_t) :: rng                          ! Random number generator
  type(mph_model_t) :: mod                        ! Model
  type(dataset_t) :: dat                          ! Dataset storage
  real(wp) :: grid_step                           ! Grid step size
  real(wp) :: beta                                ! Generic value of beta
  real(wp) :: Q                                   ! Population objective fn.
  real(wp) :: Q_n                                 ! Sample objective fn.
  integer :: i                                    ! Grid point index

  !===== Main Program ========================================================!

  ! Seed the PRNG
  call osl_rng_seed(rng, 2480129)

  ! Initialize the model
  call mph_model_init(mod, alpha, x1, x2)

  ! Generate data
  call dataset_init(dat, nn, mph_model_n_t, mph_model_dim_x)
  call mph_model_generate_data(mod, beta0, rng, dat)

  ! Open output file
  open(UNIT=lun, FILE='qmph.dat')

  ! Store population and sample objective function values at grid points
  grid_step = (grid_ub - grid_lb) / real(grid_size - 1, wp)
  do i = 1, grid_size
     beta = grid_lb + (i - 1) * grid_step
     call mph_model_population_objective(mod, beta0, beta, Q)
     call mph_model_sample_objective(mod, dat, beta, Q_n)
     write(lun, '(g17.5,1x,g17.5,1x,g17.5)') beta, Q, Q_n
  end do

  ! Close output file
  close(lun)

  ! Print gnuplot commands
  if (opt_xfig) then
     print '(a)', 'set terminal fig textspecial'
     print '(a)', 'set output "qmph.fig"'
  end if
  print '(a)', 'set size 0.8,0.8'
  print '(a)', 'set xlabel "$\\beta$"'
  print '(a)', 'plot "qmph.dat" using 1:2 with steps lt 1 title "$Q(\\beta)$",' &
       // ' "qmph.dat" using 1:3 with steps lt 3 title "$Q_n(\\beta)$"'

  ! Free used memory
  call dataset_free(dat)
  call mph_model_free(mod)
end program plotqmph
