module utility
  use osl_core, wp => osl_wp
  implicit none

  interface print_vec
     module procedure print_vec_i
     module procedure print_vec_wp
  end interface print_vec

  ! Timer variables
  integer :: count0, count1, count_rate, time
  real(wp) :: time0, time1

contains

  ! Returns .TRUE. if the interval A covers B and .FALSE. otherwise.
  function cover_p(a, b)
    real(wp), dimension(2), intent(in) :: a
    real(wp), dimension(2), intent(in) :: b
    logical :: cover_p

    cover_p = (a(1) <= b(1) .and. b(2) <= a(2))
  end function cover_p

  ! Calculates the Hausdorff distance between two intervals A and B.
  function hausdorff(a, b) result(d)
    real(wp), dimension(2), intent(in) :: a
    real(wp), dimension(2), intent(in) :: b
    real(wp) :: d
    d = maxval(abs(a - b))
  end function hausdorff

  ! Calculates the P-th quantile of a _sorted_ vector of real numbers
  ! X using interpolation:
  !
  !     quantile = (1 - \delta) x_i + \delta x_{i+1}
  !
  ! where i is floor(n * p) and \delta is n * p - i.  Thus,
  ! when P is zero, the minimum of the data is returned, when P is
  ! one, the maximum value is returned, and when P is 0.5, the median
  ! value is returned.
  function quantile(x, p)
    real(wp), dimension(:), intent(in) :: x
    real(wp),               intent(in) :: p

    real(wp) :: quantile
    integer :: n, lhs
    real(wp) :: idx, delta

    n = size(x, 1)
    idx = p * n
    lhs = int(idx)
    delta = idx - real(lhs, wp)

    if (n == 0) then
       quantile = 0.0_wp
    else if (lhs == n) then
       quantile = x(lhs)
    else
       quantile = (1.0_wp - delta) * x(lhs) + delta * x(lhs + 1)
    end if
  end function quantile

  ! Determine the lower and upper bounds of the interval, returned in
  ! SET, which maximizes the functional values Y over the grid points
  ! X up to some slackness TAU.  N denotes the sample size which is
  ! used for normalizing the functional values.
  subroutine get_max_set(x, y, n, tau, set)
    real(wp), dimension(:),       intent(in)  :: x
    real(wp), dimension(size(x)), intent(in)  :: y
    integer,                      intent(in)  :: n
    real(wp),                     intent(in)  :: tau
    real(wp), dimension(2),       intent(out) :: set

    real(wp), dimension(size(x)) :: y_norm
    integer :: i

    ! Normalize and rescale functional values.
    y_norm = sqrt(real(n, wp)) * (y - maxval(y))

    ! Determine the smallest index I such that Y(I) >= -tau.
    do i = 1, size(x)
       if (y_norm(i) >= - tau) then
          set(1) = x(i)
          exit
       end if
    end do

    ! Determine the largest index I such that Y(I) >= -tau.
    do i = size(x), 1, -1
       if (y_norm(i) >= - tau) then
          set(2) = x(i)
          exit
       end if
    end do
  end subroutine get_max_set

  ! Print REAL(WP) vector with a prefix, postfix, and custom format
  ! and separator.
  subroutine print_vec_i(pre, vec, fmt, sep, post)
    character(len=*),               intent(in) :: pre
    integer,          dimension(:), intent(in) :: vec
    character(len=*),               intent(in) :: fmt
    character(len=*),               intent(in) :: sep
    character(len=*),               intent(in) :: post
    integer :: i, n

    n = size(vec)
    write (*, '(a)', advance='no') pre
    do i = 1, n - 1
       write (*, '(' // fmt // ')', advance='no') vec(i)
       write (*, '(a)', advance='no') sep
    end do
    write (*, '(' // fmt // ')', advance='no') vec(n)
    write (*, '(a)') post
  end subroutine print_vec_i

  ! Print REAL(WP) vector with a prefix, postfix, and custom format
  ! and separator.
  subroutine print_vec_wp(pre, vec, fmt, sep, post)
    character(len=*),               intent(in) :: pre
    real(wp),         dimension(:), intent(in) :: vec
    character(len=*),               intent(in) :: fmt
    character(len=*),               intent(in) :: sep
    character(len=*),               intent(in) :: post
    integer :: i, n

    n = size(vec)
    write (*, '(a)', advance='no') pre
    do i = 1, n - 1
       write (*, '(' // fmt // ')', advance='no') vec(i)
       write (*, '(a)', advance='no') sep
    end do
    write (*, '(' // fmt // ')', advance='no') vec(n)
    write (*, '(a)') post
  end subroutine print_vec_wp

  ! Start timer
  subroutine tic()
    call system_clock(count0)
    call cpu_time(time0)
  end subroutine tic

  ! Stop timer and report elapsed wall clock and CPU time
  subroutine toc()
    call system_clock(count1, count_rate)
    call cpu_time(time1)
    time = int((count1 - count0) / real(count_rate))
    print '(a,i0,a,i0,a)', '% Elapsed time: ', &
         time, ' sec. (', int(time1 - time0), ' sec. cpu time)'
    print '(a)',    '%'
  end subroutine toc

end module utility
